@@ -2,6 +2,7 @@ import logging
 import threading
 import functools
 import confluent_kafka
+import confluent_kafka.admin
 from datetime import datetime
 from json import loads as json_loads
 from json import dumps as json_dumps
@@ -10,6 +11,7 @@ from pydantic import BaseModel
 from dataclasses import dataclass
 from collections import defaultdict, ChainMap
 from .g import _message_ctx_stack, Context
+from .const import split_props
 
 logger = logging.getLogger(__name__)
 
@@ -72,10 +74,17 @@ class TopicPartition:
 		orm_mode = True
 
 class Config(dict):
-	def __init__(self, both={}, producer={}, consumer={}):
-		super().__init__(both)
-		self.producer = ChainMap(producer, self)
-		self.consumer = ChainMap(consumer, self)
+	def __init__(self, both=None, producer=None, consumer=None):
+		if both is None:
+			both = {}
+		C,P,B = split_props(both)
+		if producer:
+			P.upadte(producer)
+		if consumer:
+			C.update(consumer)
+		super().__init__(B)
+		self.producer = ChainMap(P, self)
+		self.consumer = ChainMap(C, self)
 
 
 def run_in_thread(poll_func, daemon=True):
@@ -109,24 +118,35 @@ class Tap(object):
 	_consumer_mutex = None
 	_producer = None
 	
-	def __init__(self, config={}):
+	def __init__(self, config=None):
 		self.config = Config(config)
 		self._lock = threading.Lock()
 		self._schema = {}
 		self._handlers = defaultdict(lambda:[])
 		self._assigned = []
 		self._error_handlers = []
+		self._before_first_handler_cb_list = []
 		self.on_assign_cb_list = []
 	
 	def schema(self, topic_name):
+		'''
+		Decorator for pydantic schema class.
+		Observed message will be validated by the class.
+		'''
 		def wrapper(cls):
 			self._schema[topic_name] = cls
 			return cls
 		return wrapper
 	
 	def handler(self, topic_name, **opts):
+		'''
+		Decorator for message handler function.
+		The function will be invoked with observed message.
+
+		Function signature is:
+		func(msg:Message)
+		'''
 		def wrapper(func):
-			new_topic = not self._handlers[topic_name]
 			self._handlers[topic_name].append((func, opts))
 			return func
 		return wrapper
@@ -137,12 +157,16 @@ class Tap(object):
 			return func
 		return wrapper
 	
+	def before_first_handler(self, fn):
+		self._before_first_handler_cb_list.append(fn)
+		return fn
+	
 	def context(self):
 		return Context(self)
 	
-	def start(self, daemon=True, init_timestamp=None):
+	def start(self, daemon=True, init_timestamp=None, create_topics=False):
 		self._started = True
-		init = self.poll_prepare(timestamp=init_timestamp)
+		init = self.poll_prepare(timestamp=init_timestamp, create_topics=create_topics)
 		
 		ctrl = run_in_thread(self.poll, daemon=daemon)
 		
@@ -185,16 +209,33 @@ class Tap(object):
 		for cb in ok:
 			self.on_assign_cb_list.remove(cb)
 	
-	def poll_prepare(self, ensure_topics=None, timestamp=None):
+	def create_topics(self, topics=[]):
+		if not topics:
+			return
+		admin = confluent_kafka.admin.AdminClient(dict(self.config.producer))
+		x=admin.create_topics([confluent_kafka.admin.NewTopic(t,1,1) for t in topics])
+		for r in x.values():
+			try:
+				r.result()
+			except confluent_kafka.KafkaException as e:
+				if e.args[0].code() == confluent_kafka.KafkaError.TOPIC_ALREADY_EXISTS:
+					pass
+				else:
+					raise
+		
+
+	def poll_prepare(self, ensure_topics=None, timestamp=None, create_topics=False):
+		'''
+		ensure_topics : consumer silently does not start handling before all topics are assinged.
+		create_topics : explicitly create topics for consumer with single partition before subscription.
+		'''
 		ret = threading.Event()
 		with self._lock:
-			init = False
 			if not self._consumer:
 				self._consumer = confluent_kafka.Consumer(dict(self.config.consumer))
 				if not self.config.consumer.get("enable.auto.commit", True):
 					# producer shall barrier consumer commit in transactional api
 					self._consumer_mutex = threading.Lock()
-				init = True
 			
 			seek_topics = set()
 			if timestamp:
@@ -237,6 +278,9 @@ class Tap(object):
 			
 			current = {p.topic for p in self._consumer.assignment()}
 			if topics - current:
+				if create_topics:
+					on_server = self._consumer.list_topics().topics.keys()
+					self.create_topics(topics - current - set(on_server))
 				self._consumer.subscribe(list(topics), on_assign=self.on_assign)
 			else:
 				ret.set()
@@ -250,12 +294,22 @@ class Tap(object):
 		
 		with self.context() as ctx:
 			ctx.raw_message = msg
+			
+			try:
+				for fn in self._before_first_handler_cb_list:
+					fn()
+				
+				self._before_first_handler_cb_list.clear()
+			except:
+				logger.error("before_first_handler failed", exc_info=True)
+				return
+			
 			if msg.error():
 				for func in self._error_handlers:
 					try:
 						func()
 					except:
-						logging.error("error_handler failed", exc_info=True)
+						logger.error("error_handler failed", exc_info=True)
 			else:
 				topic = msg.topic()
 				for func,opts in self._handlers[topic]:
@@ -268,15 +322,15 @@ class Tap(object):
 					try:
 						func(m)
 					except:
-						logging.error("handler failed", exc_info=True)
+						logger.error("handler failed", exc_info=True)
 		
 		if self._consumer_mutex:
 			with self._consumer_mutex:
 				self._consumer.commit(msg)
 	
-	def map_reduce(self, topic, message=None, json=None, topic_filter=[]):
+	def map_reduce(self, topic, message=None, json=None, topic_filter=[], create_topics=True):
 		return MapReduce(
-			tap=self, topic_filter=topic_filter
+			tap=self, topic_filter=topic_filter, create_topics=create_topics
 		).map(
 			topic, message=message, json=json
 		).reduce()
@@ -286,7 +340,7 @@ class MapReduce:
 	_capture = None
 	_cb_entry = None
 	
-	def __init__(self, tap, topic_filter=[]):
+	def __init__(self, tap, topic_filter=[], create_topics=False):
 		self.tap = tap
 		self.topic_filter = topic_filter
 		
@@ -295,7 +349,10 @@ class MapReduce:
 		for t in topic_filter:
 			self.tap._handlers[t].append(self._cb_entry)
 		
-		ev = self.tap.poll_prepare(ensure_topics=topic_filter)
+		# Producer will do create the topics, we want to subscribe to that
+		# topic before produce(). Enable create_topics if you do not create
+		# topic manually.
+		ev = self.tap.poll_prepare(ensure_topics=topic_filter, create_topics=create_topics)
 		if ev:
 			while not ev.is_set():
 				if self.tap._started:
